@@ -14,6 +14,7 @@ use Devel::NYTProf::Constants qw(
 
     NYTP_SCi_CALL_COUNT NYTP_SCi_INCL_RTIME NYTP_SCi_EXCL_RTIME
     NYTP_SCi_INCL_UTIME NYTP_SCi_INCL_STIME NYTP_SCi_RECI_RTIME
+    NYTP_SCi_CALLING_SUB
 );
 
 
@@ -114,6 +115,16 @@ sub is_pmc {
 }
 
 
+sub is_perl_std_lib {
+    my $self = shift;
+    my $filename = $self->filename;
+    my $attributes = $self->profile->attributes;
+    return 1 if $filename =~ /\Q$attributes->{PRIVLIB_EXP}/;
+    return 1 if $filename =~ /\Q$attributes->{ARCHLIB_EXP}/;
+    return 0;
+}
+
+
 # should return the filename that the application used
 # when loading the file
 sub filename_without_inc {
@@ -224,12 +235,13 @@ sub dump {
             my $subs_called = $sub_call_lines->{$line};
 
             for my $subname (sort keys %$subs_called) {
-                my $sc = $subs_called->{$subname};
+                my @sc = @{$subs_called->{$subname}};
+                $sc[NYTP_SCi_CALLING_SUB] = join "|", keys %{ $sc[NYTP_SCi_CALLING_SUB] };
 
                 printf $fh "%s%s%s%s%s%s%s[ %s ]\n", 
                     $prefix, 'call', $separator,
                     $line,  $separator, $subname, $separator,
-                    join(" ", map { defined($_) ? $_ : 'undef' } @$sc)
+                    join(" ", map { defined($_) ? $_ : 'undef' } @sc)
             }
         }
 
